import os

from conf_engine.exceptions import ValueNotFound


class EnvironmentParser:
    def get_option_value(self, option: str, group: str = None):
        env_name = option.upper() if not group else group.upper() + '_' + option.upper()
        value = os.getenv(env_name)
        if value:
            return value
        else:
            raise ValueNotFound(option)





import pytest

import conf_engine.options as o


def test_string_option():
    """Test default values work and cast to string accordingly."""
    opt = o.StringOption('test_string_option')
    assert opt('string_value') == 'string_value'


def test_string_option_to_type_constructor():
    """Verify that all options are passed through to the type constructor."""
    max_length = 10
    choices = ['choice_one', 'choice_two']
    ignore_case = True
    opt = o.StringOption('test_string_option', max_length=max_length, choices=choices, ignore_case=ignore_case)
    assert opt.option_type._max_length == max_length
    assert opt.option_type._choices == choices
    assert opt.option_type._ignore_case == ignore_case


def test_number_option():
    """Test default values work and cast to int accordingly."""
    opt = o.NumberOption('test_number_option')
    assert opt('8675309') == 8675309


def test_number_option_to_type_constructor():
    """Verify that all options are passed through to the type constructor."""
    minimum = 0.38
    maximum = 0.45
    choices = [0.38, 0.44, 0.45]
    cast = float
    opt = o.NumberOption('test_number_option', minimum=minimum, maximum=maximum, choices=choices, cast=cast)
    assert opt.option_type._minimum == minimum
    assert opt.option_type._maximum == maximum
    assert opt.option_type._choices == choices
    assert opt.option_type._cast == cast


def test_boolean_option():
    """Test default values work and cast to bool accordingly."""
    opt = o.BooleanOption('test_boolean_option')
    assert opt('True') is True


def test_proper_default():
    o.NumberOption('test_string_option', default=55)


def test_improper_default():
    with pytest.raises(ValueError):
        o.NumberOption('test_string_option', default='a string')



from typing import Callable, Iterable, Union

import conf_engine.types as t


class _UndefinedDefault:
    """An object that represents a default that is not defined.  Should not be
    used outside the Config Engine module itself."""
    pass


UNDEFINED = _UndefinedDefault()


class Option:
    """
    CLass representing an option to be registered with the Configuration object.
    All Options have a subset of shared properties, name, default, type, etc.
    Inheriting classes may implement additional kwargs as appropriate to their type.
    When a default is set it will be validated using the option_type specified.
    """

    def __init__(self, name, option_type: t.Type = None, default: any = UNDEFINED):
        self.name = name
        self.option_type = option_type if option_type else t.String()
        self.default = default
        if default is not UNDEFINED:
            try:
                option_type(default)
            except ValueError as e:
                raise e

    def __str__(self):
        return self.name

    def __call__(self, value):
        return self.option_type(value)

    def __eq__(self, other: 'Option'):
        return self.name == other.name and self.option_type == other.option_type


class StringOption(Option):
    def __init__(self, *args, option_type: t.String = None, ignore_case: bool = False, max_length: int = None,
                 choices: Iterable = None, type_name: str = 'string type', **kwargs):
        if not option_type:
            kwargs['option_type'] = t.String(ignore_case=ignore_case, max_length=max_length, choices=choices,
                                             type_name=type_name)
        super().__init__(*args, **kwargs)


class NumberOption(Option):
    def __init__(self, *args, option_type: t.Number = None, minimum: Union[int, float] = None,
                 maximum: Union[int, float] = None, choices: Iterable = None,
                 type_name: str = 'number type', cast: Callable = int, **kwargs):
        if not option_type:
            kwargs['option_type'] = t.Number(minimum=minimum, maximum=maximum,
                                             choices=choices, type_name=type_name, cast=cast)
        super().__init__(*args, **kwargs)


class BooleanOption(Option):
    def __init__(self, *args, option_type=t.Boolean(), **kwargs):
        kwargs['option_type'] = option_type
        super().__init__(*args, **kwargs)



import conf_engine.types as types
import pytest


class TestString:
    def test_case_sensitive(self):
        t = types.String()
        assert t('String') != 'string'
        assert t('String') == 'String'

    def test_invalid_choices(self):
        t = types.String(choices=['string1', 'string2'])
        with pytest.raises(ValueError):
            assert t('string3')

    def test_case_sensitive_choices(self):
        t = types.String(choices=['string1', 'string2'])
        assert t('string1') == 'string1'
        with pytest.raises(ValueError):
            assert t('String2')

    def test_case_insensitive_choices(self):
        t = types.String(choices=['string1', 'string2'], ignore_case=True)
        assert t('String1') == 'string1'
        assert t('string2') == 'string2'


class TestNumber:
    def test_number(self):
        t = types.Number()
        assert t('12345') == 12345

    def test_integer(self):
        t = types.Integer()
        assert t('12345') == 12345

    def test_float(self):
        t = types.Float()
        assert t('1.2345') == 1.2345

    def test_is_integer_method(self):
        assert types.Number._is_integer('1')
        assert types.Number._is_integer(1)
        assert not types.Number._is_integer('string')

    def test_is_float_method(self):
        assert types.Number._is_float('1.5')
        assert types.Number._is_float(1.5)
        assert not types.Number._is_float('string')

    def test_is_number_method(self):
        assert types.Number._is_number('1.5')
        assert types.Number._is_number(1.5)
        assert types.Number._is_number('1.5')
        assert types.Number._is_number(1.5)
        assert not types.Number._is_number('string')

    def test_minimum(self):
            t = types.Number(minimum=10)
            assert t('10') == 10
            with pytest.raises(ValueError):
                assert t('9')

    def test_maximum(self):
        t = types.Number(maximum=10)
        assert t('10') == 10
        with pytest.raises(ValueError):
            assert t('11')

    def test_choices(self):
        choices = ['1', '2', '3']
        t = types.Number(choices=choices)
        for choice in choices:
            assert t(choice) == int(choice)

    def test_invalid_choice(self):
        t = types.Number(choices=['1', '2'])
        with pytest.raises(ValueError):
            assert t('3')

    def test_minimum_over_maximum(self):
        with pytest.raises(ValueError):
            assert types.Number(maximum=1, minimum=10)


class TestBoolean:
    def test_true_values(self):
        t = types.Boolean()
        assert t(True)
        assert t('yes')
        assert t('true')
        assert t('True')
        assert t(1)
        assert t('1')
        assert t('on')

    def test_false_values(self):
        t = types.Boolean()
        assert not t(False)
        assert not t('no')
        assert not t('No')
        assert not t('False')
        assert not t(0)
        assert not t('0')
        assert not t('off')

    def test_not_boolean(self):
        t = types.Boolean()
        with pytest.raises(ValueError):
            assert t('not boolean')




import conf_engine.parsers.ini_file as ini_file

from conf_engine.tests.conftest import test_ini_directory


def test_config_cli_args_with_no_args(monkeypatch):
    path_opts = ini_file.INIFileParser._register_paths()
    assert path_opts['config_dirs'] == ['./']
    assert path_opts['config_files'] == []


def test_config_cli_args_with_files(monkeypatch):
    monkeypatch.setattr('sys.argv', ['program', '--config-dir', '/path/one', '--config-dir', '/path/two'])
    path_opts = ini_file.INIFileParser._register_paths()
    assert path_opts['config_dirs'] == ['/path/one', '/path/two']
    assert path_opts['config_files'] == []


def test_config_cli_args_with_dirs(monkeypatch):
    monkeypatch.setattr('sys.argv', ['program', '--config-file', 'file1.ini', '--config-file', 'file2.ini'])
    path_opts = ini_file.INIFileParser._register_paths()
    assert path_opts['config_dirs'] == []
    assert path_opts['config_files'] == ['file1.ini', 'file2.ini']


def test_load_configs(monkeypatch):
    monkeypatch.setattr('sys.argv', ['program', '--config-file', './test.ini', '--config-dir', './test_ini_files'])
    ini_file.INIFileParser()


def test_ignore_unknown_args(monkeypatch):
    monkeypatch.setattr('sys.argv', ['program', '--unknown_arg'])


def test_get_config_values(test_ini_directory, monkeypatch):
    monkeypatch.chdir(test_ini_directory)
    monkeypatch.setattr('sys.argv', ['program', '--config-file', './test.ini', '--config-dir', './types'])
    ifp = ini_file.INIFileParser()
    assert ifp.get_option_value('default_option') == 'default_value'
    assert ifp.get_option_value('integer', 'numbers') == '12345'
    assert ifp.get_option_value('boolean_no', 'booleans') == 'no'



import argparse
import configparser
import os

from conf_engine.exceptions import ValueNotFound


class INIFileParser:
    def __init__(self):
        self._path_opts = self._register_paths()

    @staticmethod
    def _register_paths():
        parser = argparse.ArgumentParser()
        parser.add_argument('--config-file',
                            type=str,
                            action='append',
                            help="Configuration file path.")
        parser.add_argument('--config-dir',
                            type=str,
                            action='append',
                            help="Directory containing configuration files to be parsed.")
        args, _ = parser.parse_known_args()
        return {
            'config_files': args.config_file or [],
            'config_dirs': ['./'] if not args.config_dir and not args.config_file else args.config_dir or []
        }

    def get_option_value(self, option: str, group: str = None):
        if not group:
            group = 'DEFAULT'
        files = self._path_opts['config_files'] + self._get_ini_files_from_dirs()
        ini_cfg = configparser.ConfigParser()
        ini_cfg.read(files)
        if group in ini_cfg and option in ini_cfg[group]:
            return ini_cfg[group][option]
        else:
            raise ValueNotFound(option)

    def _get_ini_files_from_dirs(self) -> list:
        files = []
        for directory in self._path_opts['config_dirs']:
            d = directory if directory.endswith('/') else directory + '/'
            files += [d + file for file in os.listdir(d) if file.endswith('.ini')]
        return files



import pytest

import conf_engine.configuration


@pytest.fixture
def test_ini_directory(request):
    path = ''
    for part in request.path.parts[1:]:
        path += '/' + part
        if part == 'conf_engine':
            return path + '/tests/test_ini_files'


@pytest.fixture
def test_config(monkeypatch, cli_opts: [str] = None, ):
    cli_opts = cli_opts or ['program']
    if cli_opts[0].startswith('-'):
        cli_opts.insert(0, 'program')

    monkeypatch.setattr('sys.argv', cli_opts)
    return conf_engine.configuration.Configuration()



class ConfigError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class UnregisteredOption(ConfigError):
    def __str__(self):
        return f"No option is registered with name: {self.message}"

class UnregisteredGroup(ConfigError):
    def __str__(self):
        return f"No group is registered with name: {self.message}"


class DuplicateOptionError(ConfigError):
    def __str__(self):
        return f"Option or group is already registered: {self.message}"


class ValueNotFound(ConfigError):
    def __str__(self):
        return f"Could not find a value for {self.message}."


import pytest


def test_register_option(test_ini_directory, test_config, monkeypatch):
    monkeypatch.chdir(test_ini_directory)
    monkeypatch.setattr('sys.argv', ['program', '--config-file', './test.ini'])
    from conf_engine.options import StringOption

    test_config.register_option(StringOption('default_option'))
    assert test_config.default_option == 'default_value'


def test_register_options(test_ini_directory, test_config, monkeypatch):
    monkeypatch.chdir(test_ini_directory)
    monkeypatch.setattr('sys.argv', ['program', '--config-file', './test.ini'])

    from conf_engine.options import StringOption, NumberOption
    options = [
        StringOption('default_option'),
        StringOption('test_option_two'),
        NumberOption('test_int'),
    ]
    test_config.register_options(options)
    assert test_config.default_option == 'default_value'
    assert test_config.test_option_two == 'test_two'
    assert test_config.test_int == 12345


def test_register_options_in_group(test_ini_directory, test_config, monkeypatch):
    monkeypatch.chdir(test_ini_directory)
    monkeypatch.setattr('sys.argv', ['program', '--config-file', './types/booleans.ini'])

    from conf_engine.options import BooleanOption
    options = [
        BooleanOption('boolean_true'),
        BooleanOption('boolean_yes'),
        BooleanOption('boolean_false'),
        BooleanOption('boolean_no')
    ]
    test_config.register_options(options, 'booleans')
    assert test_config.booleans.boolean_true
    assert test_config.booleans.boolean_yes
    assert not test_config.booleans.boolean_false
    assert not test_config.booleans.boolean_no


def test_register_same_option_name_with_different_params(test_config):
    from conf_engine.options import StringOption, NumberOption
    from conf_engine.exceptions import DuplicateOptionError

    test_config.register_option(StringOption('default_option'))
    with pytest.raises(DuplicateOptionError):
        test_config.register_option(NumberOption('default_option'))


def test_register_same_option_name_with_same_params(test_config):
    from conf_engine.options import StringOption

    test_config.register_option(StringOption('default_option'))
    test_config.register_option(StringOption('default_option'))


def test_default_options(test_ini_directory, test_config, monkeypatch):
    monkeypatch.chdir(test_ini_directory)
    monkeypatch.setattr('sys.argv', ['program', '--config-file', './test.ini'])

    from conf_engine.options import BooleanOption, NumberOption, StringOption
    options = [
        StringOption('default_option', default='This should not return the default.'),
        NumberOption('test_option_default', default=100),
        BooleanOption('test_bool_option', default=True),
        BooleanOption('test_bool_false_option', default=False),
        StringOption('test_none_default_option', default=None)
    ]
    test_config.register_options(options)
    assert test_config.test_bool_option
    assert not test_config.test_bool_false_option
    assert test_config.default_option == 'default_value'
    assert test_config.test_option_default == 100
    assert test_config.test_none_default_option is None


def test_option_precedence(test_ini_directory, test_config, monkeypatch):
    monkeypatch.chdir(test_ini_directory)
    monkeypatch.setattr('sys.argv', ['program', '--config-file', './test.ini'])
    monkeypatch.setenv('DEFAULT_OPTION', 'env_value')

    from conf_engine.options import StringOption
    options = [
        StringOption('default_option', default='This should not return the default.')
    ]
    test_config.register_options(options)
    assert test_config.default_option == 'env_value'


def test_option_value_caching(test_config, monkeypatch):
    monkeypatch.setenv('DEFAULT_OPTION', 'env_value')
    from conf_engine.options import StringOption
    options = [
        StringOption('default_option', default='This should not return the default.')
    ]
    test_config.register_options(options)
    assert not test_config._get_group(None)._option_value_cached('default_option')
    _ = test_config.default_option
    assert test_config._get_group(None)._option_value_cached('default_option')
    assert test_config._get_group(None)._get_option_value_from_cache('default_option') == 'env_value'


def test_option_value_cache_flush(test_config, monkeypatch):
    monkeypatch.setenv('OPG_STR_OPTION', 'opt_value')
    from conf_engine.options import StringOption
    options = [
        StringOption('str_option')
    ]
    test_config.register_options(options, 'opg')
    assert not test_config._get_group('opg')._option_value_cached('str_option')
    _ = test_config.opg.str_option
    assert test_config._get_group('opg')._option_value_cached('str_option')
    test_config.flush_cache()
    assert not test_config._get_group('opg')._option_value_cached('str_option')
    _ = test_config.opg.str_option
    assert test_config._get_group('opg')._option_value_cached('str_option')



import logging

import conf_engine.exceptions as cfg_exc
import conf_engine.parsers as parsers

from conf_engine.options import Option, UNDEFINED

REGISTERED_PARSERS = [
    parsers.EnvironmentParser,
    parsers.INIFileParser,
]


class ConfigGroup:
    def __init__(self, name, cache: bool = True):
        """
        A collection of related configuration options.
        :param name:
        :param cache: When True (default) this ConfigGroup will store
                      values after they are read from configuration.
        """
        self._name = name
        self._cache = cache
        self._opt_cache = {}
        self._value_cache = {}

    def __getattr__(self, item: str):
        return self._get_option(item)

    def __contains__(self, item):
        return item in self._opt_cache

    def _get_option(self, option: str):
        if option in self._opt_cache:
            return self._get_option_value(self._opt_cache[option], self._name)
        raise cfg_exc.UnregisteredOption(option)

    def _cache_option_value(self, name, value):
        self._value_cache[name] = value

    def _get_option_value_from_cache(self, name):
        return self._value_cache[name]

    def _option_value_cached(self, name):
        return name in self._value_cache if self._cache else False

    def _get_option_value(self, option: Option, group):
        if self._option_value_cached(option.name):
            return self._get_option_value_from_cache(option.name)
        else:
            return self._get_option_value_from_source(option, group)

    def _get_option_value_from_source(self, option: Option, group):
        for parser in REGISTERED_PARSERS:
            try:
                value = parser().get_option_value(option.name, group)
                # Validate the value is correctly formatted.
                value = option.option_type(value)
                # Store the value in the value cache.
                self._cache_option_value(option.name, value)
                # The first parser in registered parsers list
                # should take precedence, so we return early.
                return value
            except cfg_exc.ValueNotFound:
                continue
            except Exception as e:
                logging.exception(e)
                raise e

        if option.default is not UNDEFINED:
            return option.option_type(option.default)
        # If we get here, then we've not found the value.
        raise cfg_exc.ValueNotFound(option.name)

    def flush_cache(self, name: str = None):
        """
        Flush the value cache and read from configuration source on
        next access.
        :param name: If name is provided, only the value for the named
        option is flushed.
        :return:
        """
        if name:
            self._value_cache.pop(name)
        else:
            self._value_cache = {}

    def register_options(self, options: [Option]):
        for option in options:
            self.register_option(option)

    def register_option(self, option: Option):
        if option.name in self._opt_cache and option != self._opt_cache[option.name]:
            raise cfg_exc.DuplicateOptionError(option.name)
        self._opt_cache[option.name] = option


class Configuration:
    def __init__(self, cache: bool = True):
        """
        Configuration object that represents the configuration of the application.
        :param cache: When True (default) Config Engine will read the value
                      of the option from the configuration source once, and then
                      store the value for subsequent lookups.  If False, then
                      the value is not stored and is always read from the
                      configuration source.  When set here, ConfigEngine will
                      set pass this along to auto created configuration groups.
        """
        self._cache = cache
        self._group_cache = {None: ConfigGroup(None)}

    def __getattr__(self, item):
        return self._get_group(item)

    def __contains__(self, item):
        return item in self._group_cache

    def register_options(self, options: [Option], group: str = None):
        """
        Register bulk options with the config.
        :param options: List of options.
        :param group: Group name to which options are added.
        :return:
        """

        for option in options:
            self.register_option(option, group=group)

    def flush_cache(self):
        """
        Signal all configuration groups to flush their cache and
        read from configuration source on next access.
        """
        for _, group in self._group_cache.items():
            group.flush_cache()

    def register_option(self, option: Option, group: str = None, create_group: bool = True):
        """
        Register options with the config.  If group is specified, the options are
        added to the option group, otherwise options are registered to the base object.
        :param option: Option to register.
        :param group: Group name to which the option is added.
        :param create_group: Create the group if not already registered.
        :return:
        """
        if group and group not in self._group_cache:
            if create_group:
                self._group_cache[group] = ConfigGroup(group, cache=self._cache)
            else:
                raise cfg_exc.UnregisteredGroup(group)
        self._group_cache[group].register_option(option)

    @property
    def registered_parsers(self):
        return self._registered_parsers

    def _get_group(self, group: str):
        """
        Get group by its name as called during attribute access against the
        configuration object.  If the option cannot be found an UnregisteredOption
        error will be raised.

        If the option matches a group name, the group object is returned and the
        subsequent attribute access is handled by the group object.
        """
        # If we find the option name in our cache, that means what was passed to
        # __getattr__() is the group value.  We return that and have the group object
        # perform the call to _get_option_value().  Otherwise, we return the default
        # group object

        if group in self._group_cache:
            return self._group_cache[group]
        else:
            return getattr(self._group_cache[None], group)



import os
import setuptools

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

here = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(here+'/conf_engine/__version__.py', 'r') as f:
    exec(f.read(), about)

install_requires = [
]

test_requires = install_requires + [
    'pytest'
]

setuptools.setup(
    name="conf_engine",
    version=about['__version__'],
    author="Ken Vondersaar",
    author_email="kvondersaar@connectria.com",
    description="A python module for unified application configuration.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/Connectria/conf-engine",
    project_urls={
        "Bug Tracker": "https://github.com/Connectria/conf-engine",
        "Documentation": "https://connectria.github.io/conf-engine",
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "./"},
    packages=setuptools.find_packages(where='./'),
    python_requires=">=3.7",
    install_requires=install_requires,
    test_requires=test_requires,
    test_suite='pytest',
    exclude_package_data={'': ['*/tests/*']},
)


import conf_engine.parsers.env as env


def test_get_env_var(monkeypatch):
    monkeypatch.setenv('TEST_VAR', 'test_var_value')
    assert env.EnvironmentParser().get_option_value('test_var') == 'test_var_value'


def test_get_env_group_var(monkeypatch):
    monkeypatch.setenv('TESTGROUP_TEST_VAR', 'test_var_value')
    assert env.EnvironmentParser().get_option_value('test_var', 'testgroup') == 'test_var_value'



from typing import Callable, Iterable, Union


class Type:
    def __init__(self, type_name: str = 'unknown type'):
        self.type_name = type_name

    def __call__(self, value):
        """
        Value is passed to the type object for the object to verify that the value
        matches the parameters defined for the type.
        :param value:
        :return:
        """
        return True

    def __eq__(self, other: 'Type'):
        return self.type_name == other.type_name


class Number(Type):
    CAST_OPTIONS = [int, float, complex]

    def __init__(self, minimum: Union[int, float] = None, maximum: Union[int, float] = None,
                 choices: Iterable = None, type_name: str = 'number type', cast: Callable = int):
        if minimum is not None and maximum is not None and maximum < minimum:
            raise ValueError("Minimum is greater than maximum.")
        if choices:
            invalid_choices = [x for x in choices if not self._is_number(x)]
            if invalid_choices:
                raise ValueError(f"Choices {invalid_choices} are not numeric.")

        if cast and cast not in self.CAST_OPTIONS:
            raise ValueError(f"Cast must be a numeric type {self.CAST_OPTIONS}")
        self._cast = cast
        self._minimum = minimum
        self._maximum = maximum
        self._choices = [self._cast(choice) for choice in choices] if choices else []
        super().__init__(type_name=type_name)

    def __call__(self, value):
        try:
            value = self._cast(value)
        except ValueError:
            raise ValueError(f"Value {value} is not correct type {self._cast}")
        if self._maximum and value > self._maximum:
            raise ValueError(f"Value {value} exceeds the maximum value of {self._maximum}")
        if self._minimum and value < self._minimum:
            raise ValueError(f"Value {value} is less than the minimum value of {self._minimum}")
        if self._choices and value not in self._choices:
            raise ValueError(f"Valid values are {self._choices}, but found {value}.")
        return value

    @staticmethod
    def _is_float(value):
        try:
            float(value)
            return True
        except ValueError:
            return False

    @staticmethod
    def _is_integer(value):
        try:
            int(value)
            return True
        except ValueError:
            return False

    @staticmethod
    def _is_number(value):
        return Number._is_integer(value) or Number._is_float(value)


class Integer(Number):
    def __init__(self, minimum: int = None, maximum: int = None,
                 choices: Iterable = None, type_name: str = 'number type',
                 cast: Callable = int):
        super().__init__(minimum=minimum, maximum=maximum, choices=choices, type_name='integer type', cast=cast)


class Float(Number):
    def __init__(self, minimum: int = None, maximum: int = None,
                 choices: Iterable = None, type_name: str = 'number type',
                 cast: Callable = float):
        super().__init__(minimum=minimum, maximum=maximum, choices=choices, type_name='float type', cast=cast)


class String(Type):
    def __init__(self, ignore_case: bool = False, max_length: int = None,
                 choices: Iterable = None, type_name: str = 'string type'):
        self._ignore_case = ignore_case
        self._max_length = max_length
        self._choices = [x.lower() for x in choices] or [] if ignore_case and choices else choices or []
        super().__init__(type_name=type_name)

    def __call__(self, value):
        value = value.lower() if self._ignore_case else value
        if self._max_length and len(value) > self._max_length:
            raise ValueError(f"Value {value} is longer than {self._max_length}.")
        if self._choices and value not in self._choices:
            raise ValueError(f"Valid values are {self._choices}, but found {value}.")
        return value


class Boolean(Type):
    TRUE = ['true', '1', 'on', 'yes']
    FALSE = ['false', '0', 'off', 'no']

    def __init__(self, type_name: str = 'boolean type'):
        super().__init__(type_name=type_name)

    def __call__(self, value):
        value = str(value)
        if isinstance(value, bool):
            return value
        if value.lower() in self.TRUE:
            return True
        if value.lower() in self.FALSE:
            return False

        raise ValueError(f"Unexpected boolean value: {value}.")
